
import fileinput
import stanza
from stanza.models.common.doc import Document
import json
import sys
import pandas as pd
import networkx
import matplotlib.pyplot as plt
from bokeh.io import output_notebook, show, save
from bokeh.models import Range1d, Circle, ColumnDataSource, MultiLine, EdgesAndLinkedNodes, NodesAndLinkedEdges, LabelSet
from bokeh.plotting import figure
from bokeh.plotting import from_networkx
from bokeh.palettes import Blues8, Reds8, Purples8, Oranges8, Viridis8, Spectral8
from bokeh.transform import linear_cmap

# Dateiname wird dem Script als Argument mitgegeben:
corpusfile = sys.argv[1]

print("Lade Datei ", corpusfile)

# Korpus laden und in ein Stanza-Objekt umwandeln:
f = open(corpusfile, "r", encoding="utf8")
doc = Document(json.load(f))

combined_list = []

# Satz für Satz prozessieren:
for sent in doc.sentences:
	source = []
	target = []
	# Token für Token prozessieren:
	for token in sent.tokens:
		if token.ner == "S-LOC": # Wenn Toponym
			source.append(token.text)
		if token.words[0].upos == "NOUN": # Wenn Nomen (Achtung: 1 Token kann mehrere Wörter enthalten)
			target.append(token.words[0].lemma)
	# alle Kombinationen von source/targets im gleichen Satz erzeugen:
	for i in source:
		for j in target:
			combined_list.append([i,j])

# Dateframe (Tabelle) erstellen:
df = pd.DataFrame(combined_list, columns=['source','target'])
# Gruppieren und gleiche Kombinationen auszählen:
res = df.groupby(['source','target']).size().reset_index(name='count')


# Filtermöglichkeit:
#res = res[res['count'] > 2]

# Netzwerk berechnen:
G = networkx.from_pandas_edgelist(res, 'source', 'target', 'count')

# Degree berechnen:
degrees = dict(networkx.degree(G))
networkx.set_node_attributes(G, name='degree', values=degrees)

# Degree anpassen, dass auch kleine Knoten dargestellt werden:
number_to_adjust_by = 5
adjusted_node_size = dict([(node, degree+number_to_adjust_by) for node, degree in networkx.degree(G)])
networkx.set_node_attributes(G, name='adjusted_node_size', values=adjusted_node_size)


#Choose attributes from G network to size and color by — setting manual size (e.g. 10) or color (e.g. 'skyblue') also allowed
size_by_this_attribute = 'adjusted_node_size'
color_by_this_attribute = 'adjusted_node_size'

#Pick a color palette — Blues8, Reds8, Purples8, Oranges8, Viridis8
color_palette = Blues8

#Choose a title!
title = 'Netzwerkvisualisierung'

#Establish which categories will appear when hovering over each node
HOVER_TOOLTIPS = [
       ("Wort", "@index"),
        ("Degree", "@degree")
]

#Create a plot — set dimensions, toolbar, and title
plot = figure(tooltips = HOVER_TOOLTIPS,
              tools="pan,wheel_zoom,save,reset", active_scroll='wheel_zoom',
            x_range=Range1d(-10.1, 10.1), y_range=Range1d(-10.1, 10.1), title=title)

#Create a network graph object
# https://networkx.github.io/documentation/networkx-1.9/reference/generated/networkx.drawing.layout.spring_layout.html\
network_graph = from_networkx(G, networkx.spring_layout, scale=10, center=(0, 0))

#Set node sizes and colors according to node degree (color as spectrum of color palette)
minimum_value_color = min(network_graph.node_renderer.data_source.data[color_by_this_attribute])
maximum_value_color = max(network_graph.node_renderer.data_source.data[color_by_this_attribute])
network_graph.node_renderer.glyph = Circle(size=size_by_this_attribute, fill_color=linear_cmap(color_by_this_attribute, color_palette, minimum_value_color, maximum_value_color))

#Set edge opacity and width
network_graph.edge_renderer.glyph = MultiLine(line_alpha=0.5, line_width=1)

plot.renderers.append(network_graph)

#Add Labels
#x, y = zip(*network_graph.layout_provider.graph_layout.values())
#node_labels = list(G.nodes())
#source = ColumnDataSource({'x': x, 'y': y, 'name': [node_labels[i] for i in range(len(x))]})
#labels = LabelSet(x='x', y='y', text='name', source=source, background_fill_color='white', text_font_size='10px', background_fill_alpha=.7)
#plot.renderers.append(labels)


show(plot)
#save(plot, filename=f"{title}.html")