"""
Skript um Topic-Modeling auf Dokumenten auszuführen und eine html-Visualisierung zum Modell zu erstellen.

Input:
    - Datei, welche ein Dokument pro Zeile enthält
    optionaler Input:
    - beliebig viele ganze Zahlen, die der Anzahl berechneten Topics entsprechen. Getrennt durch einen Abstand.
    - beliebig viele Werte zwischen 0 und 1, welche dem oberen Limit an Auftreten in den Dokumenten entsprechen,
        ab wann ein Wort zum Stopword wird. Bsp. 0.8: Sobald ein Wort in mehr als 0.8 der Dokumente vorkommt,
        wird es ignoriert. Die Werte sind mit Abstand dazwischen einzugeben, also z.B. 0.1 0.5 0.85

Output:
    - .csv Dateien, die den verschiedenen Models nach Angabe der Parameter entsprechen, 
        einmal mit den Wörtern pro Topic, einmal mit den Topic-Distributionen pro Dokument
    - .html Dateien, die eine Visualisierung der Topics enthalten und sich im Browser öffnen lassen


Quelle neben der Dokumentation von radimrehurek:
https://medium.com/analytics-vidhya/topic-modeling-using-gensim-lda-in-python-48eaa2344920
https://neptune.ai/blog/pyldavis-topic-modelling-exploration-tool-that-every-nlp-data-scientist-should-know
https://stackoverflow.com/questions/43014259/display-formatter-attribute-error-in-python
"""

from gensim.corpora.dictionary import Dictionary
from gensim.models.ldamodel import LdaModel
from gensim.models.coherencemodel import CoherenceModel
import argparse
import pyLDAvis
import pyLDAvis.gensim_models

from typing import List, TextIO, Generator


def parse_arguments():
    my_parser = argparse.ArgumentParser(
        description="Erstellt ein Topic Modeling, entweder mit default Werten oder mit eigenen Anzahl Topics und Stopwords-limit"
    )
    my_parser.add_argument(
        "in_file",
        type=argparse.FileType("r", encoding="utf-8"),
        help="Input-Datei: Lemmatisierte Dokumente",
    )
    my_parser.add_argument(
        "-nt",
        "--number_topics",
        type=int,
        nargs="+",
        default=[7, 15],
        help="optionale Angabe verschiedener Anzahlen von Topics, default ist 7 und 15",
    )
    my_parser.add_argument(
        "-ul",
        "--upper_limits",
        type=float,
        nargs="+",
        default=[0.5, 0.8],
        help="optionale Angabe mehrerer Limiten für die Anzahl Dokumente, "
        "in denen ein Wort vorkommen kann, bevor es zum stopword wird."
        "wobei 1 = jedes Dokument, 0 = kein Dokument. Defaultwerte sind 0.5 und 0.8",
    )
    return my_parser.parse_args()


def lines_generator(infile: TextIO) -> Generator[List[str], None, None]:
    """Generiert eine Liste mit dem Inhalt der einzelnen Zeilen/Dokumente der Input-Datei.

    Args:
        infile (TextIO): Input-Datei mit lemmatisierten Dokumenten

    Yields:
        List[str]: Liste mit tokens eines Dokuments
    """

    for line in infile:
        word_list = line.strip().split(" ")
        yield word_list


def html_visualisation(
    lda_model: LdaModel, corpus, Dict_id_words: Dictionary, nr: int, limit: float
) -> None:
    """
    Erstellt eine html-Visualisierung des Topic Models und speichert sie als html-Datei ab.
    """
    visualisation = pyLDAvis.gensim_models.prepare(
        lda_model, corpus, Dict_id_words, mds="mmds", sort_topics=False
    )
    pyLDAvis.save_html(
        visualisation,
        "tm_vis_topics" + str(nr) + "_upperlimit" + str(limit) + ".html",
    )
    print(
        "Das file tm_vis_topics"
        + str(nr)
        + "_upperlimit"
        + str(limit)
        + ".html wurde ausgegeben"
    )


def produce_outfiles(
    topics: list, lda_model: LdaModel, corpus, nr: int, limit: float
) -> None:
    """
    Erstellt zwei .csv Dateien mit den Wörtern pro Topic und den Topic-Distributionen pro Dokument.
    """

    with open(
        "tm_topic_content_topics" + str(nr) + "upperlimit" + str(limit) + ".csv",
        "w",
    ) as open_out_file1:
        open_out_file1.write("Topic, Words, \n")
        # schreibt pro Topic eine Zeile mit den num_words wichtigsten Wörtern in ein csv-file
        for curr_topic in topics:
            curr_words_in_topic = curr_topic[1].split("+")
            outline_for_topic = "Topic " + str(curr_topic[0]) + ", "
            for word in curr_words_in_topic:
                outline_for_topic += word.split('"')[1] + " "
            open_out_file1.write(outline_for_topic + "\n")

    with open(
        "tm_topics_per_doc_topics" + str(nr) + "upperlimit" + str(limit) + ".csv", "w"
    ) as open_out_file2:
        # erstellt eine Liste mit den Topic-Distributionen pro Dokument
        topic_per_doc = lda_model.get_document_topics(corpus)
        open_out_file2.write("Document Number, Topic distribution, \n")
        # schreibt für jedes Dokument die Topic-Distribution in eine Zeile
        for index, topics_per_doc in enumerate(topic_per_doc):
            outline_per_doc = "Document " + str(index) + ", "
            for topic in topics_per_doc:
                outline_per_doc += str(topic) + " "
            open_out_file2.write(f"{outline_per_doc}, \n")
        print(
            "Das file tm_topics_per_doc_topics"
            + str(nr)
            + "upperlimit"
            + str(limit)
            + ".csv wurde ausgegeben"
        )


def topic_modeling(
    infile: TextIO, number_topics: int, upper_limits: List[float]
) -> None:
    """Berechnet Topic Modeling mit den angegebenen Parametern.

    Args:
        infile (TextIO): Input-Datei mit einem Dokument pro Zeile
        number_topics (int): Anzahl Topics, die berechnet werden sollen
        upper_limits (List[float]): Liste mit den oberen Limits für die
                     Anzahl Dokumente, in denen ein Wort vorkommen darf

    Returns:
        None
    """
    stop_word_upper_limits = upper_limits
    nr_of_topics = number_topics
    Dict_id_words = {}
    all_documents = []

    documents = lines_generator(infile)
    print("reading input...")

    for doc in documents:
        all_documents.append(doc)

    Dict_id_words = Dictionary(all_documents)

    for limit in stop_word_upper_limits:
        # entfernt Wörter, die in mehr als limit der Dokumente vorkommen
        Dict_id_words.filter_extremes(
            no_below=1, no_above=limit, keep_n=None, keep_tokens=None
        )
        # erstellt ein corpus mit den relevanten Dokumenten
        corpus = [Dict_id_words.doc2bow(text) for text in all_documents]
        print(limit)
        print("Building model ....")

        for nr in nr_of_topics:
            # erstellt ein LDA model für die Anzahl Topics
            lda_model = LdaModel(
                corpus=corpus,
                id2word=Dict_id_words,
                num_topics=nr,
                random_state=100,
                iterations=100,
                update_every=1,
                chunksize=100,
                passes=10,
                alpha="auto",
                per_word_topics=True,
            )

            html_visualisation(lda_model, corpus, Dict_id_words, nr, limit)

            # erstellt eine liste von (nr_topics) Topics mit den num_words wichtigsten Wörtern pro Topic aus
            topics = lda_model.show_topics(num_topics=nr, num_words=20)

            produce_outfiles(topics, lda_model, corpus, nr, limit)

        # füllt den Dictionary wieder auf, damit die nächste Iteration
        # mit dem ursprünglichen Dictionary beginnt
        Dict_id_words = Dictionary(all_documents)


def main():
    args = parse_arguments()
    topic_modeling(args.in_file, args.number_topics, args.upper_limits)


if __name__ == "__main__":
    main()
