#!/usr/bin/env python3
"""Wandelt .vrt-Files in 1-Text-pro-Zeile Files um.

Mit diesem Skript kann der .vrt-Output von Promethia oder Befehlen wie `cwb-decode -C
<CORPUS> -ALL` in ein Format umgewandelt werden, wie es von `topic-_modeling.py` in
diesem Verzeichnis verarbeitet werden kann.

Das schlussendliche Format enthält ein Dokument/Text pro Zeile. Die Tokens sind mit
Leerzeichen getrennt. Mit der Kommandozeilenoption `-c` (Column) kann angegeben werden,
welche Dimension des Tokens ausgegeben werden soll. Dazu muss im Urspungsfile (.vrt)
nachgeschaut werden, in welcher Token-Spalte sich die relevante Dimension befindet. Ist
es die erste Spalte, wird das Skript mit `-c 0` aufgerufen (Von 0 an zählen).
"""

import argparse
from typing import Generator, List, TextIO


def parse_arguments():
    my_parser = argparse.ArgumentParser(
        description="Transforms the output of cwb-decode to a tabular format for further processing."
    )

    my_parser.add_argument(
        "input",
        action="store",
        type=argparse.FileType("r", encoding="utf-8"),
        help="Output of cwb-decode",
    )
    my_parser.add_argument(
        "output",
        action="store",
        type=argparse.FileType("w", encoding="utf-8"),
        help="File with a tokenized Document per line",
    )

    my_parser.add_argument(
        "-c",
        "--column",
        action="store",
        type=int,
        help="Which column of the token rows in the .vrt-file should be considered",
        required=True,
    )

    return my_parser.parse_args()


def document_generator(infile: TextIO, column: int) -> Generator[List[str], None, None]:
    """Generates token-lists corresponding to a document/text.

    Args:
        infile (TextIO): Opened File to read from.
        column (int): Column number (0-indexed) of the relevant token dimension.

    Yields:
        List[str]: A list of tokens.
    """

    tokens = []

    for line in infile:

        line = line.strip()

        if line == "":
            continue

        # Is the line part of a structural attribute
        elif line.startswith("<") and line.endswith(">"):
            if line.startswith("<text"):
                tokens = []
            elif line.startswith("</text>"):
                yield tokens
            # Structural attributes apart from `text` are ignored
            else:
                continue

        # Line represents a token
        else:
            token_dimensions = line.split("\t")
            tokens.append(token_dimensions[column])


def main():
    args = parse_arguments()
    documents = document_generator(args.input, column=args.column)

    outfile = args.output

    first_doc_str = " ".join(next(documents))
    outfile.write(first_doc_str)

    for doc in documents:
        outfile.write("\n")
        doc_str = " ".join(doc)
        outfile.write(doc_str)


if __name__ == "__main__":
    main()
