"""
Skript um Nearest Neighbors eines Wortes bzw. dessen Vektors aus einem Word-Embedding Modelles zu visualisieren.

Dieser Code stammt bis auf die Variante mit 2D und minimalen Anpassungen aus folgendem Git Repo von Niclas Bodenmann:
https://gitlab.uzh.ch/niclaslinus.bodenmann/we-and-plotly/-/blob/master/Playground.ipynb

Input:
    - Datei im word_vectors Format von Gensim, welche das Word-Embedding Modell enthält
    - Angabe der gewünschen Dimensionen, in denen visualisiert werden soll: '2d' oder '3d' eingeben

Benötigte Anpassungen im Code: (in der Funktion we_pca_vis am Schluss des Skripts)
    - Die Wörter (beliebig viele), deren Umfeld untersucht werden soll
    - Anzahl der Nearest Neighbors, die pro Wort angezeigt werden sollen

Output:
    - Visualisierung, die sich automatisch im Browser öffnen sollte und von da als png heruntergeladen werden kann
"""

import argparse
from gensim.models import Word2Vec, KeyedVectors
import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
import plotly.express as px
import plotly.graph_objs as go

dims = ""

def nearest_neighbours(w, k, model):
    # verpackt die NN in ein Array
    most_similar = model.most_similar(positive=[w], topn=k)
    nearest_neighbours = {}
    nearest_neighbours[w] = model[w]
    for x in most_similar:
        nearest_neighbours[x[0]] = model[x[0]]
    return nearest_neighbours

def sk_PCA(df):
    # performs PCA on array 'df'
    if dims == '2d':
        pca = PCA(n_components= 2)
    elif dims == '3d':
        pca = PCA(n_components= 3)
    else:
        print("keine gültige Dimension gewählt: '2d' oder '3d' eingeben")
        quit()
    return pca.fit_transform(df)


def return_trace(df, word):
    if dims == '2d':
        trace = go.Scatter(
            name=word,
            x=df[0],
            y=df[1],
            text=df.index,
            mode="markers+text",
            textfont=dict(
                color=["crimson"]),
            marker=dict(
                size=3),
            hoverinfo="skip"
        )
    elif dims == '3d':
        trace = go.Scatter3d(
            name=word,
            x=df[0],
            y=df[1],
            z=df[2],
            text=df.index,
            mode="markers+text",
            textfont=dict(
                color=["crimson"]),
            marker=dict(
                size=3),
            hoverinfo="skip"
        )
    return trace


def figure_with_multiple_traces(tracelist):
    fig = go.Figure()

    buttonlist = []
    for i in range(len(tracelist)):
        trace = tracelist[i]
        if i == 0:
            trace.visible = True
        else:
            trace.visible = False
        button = dict(
            label=trace.name,
            method="update",
            args=[{"visible": [True if n == i else False for n in range(len(tracelist))]},
                  {"title": trace.name},
                  {"showlegend": True}])
        buttonlist.append(button)

    for trace in tracelist:
        fig.add_trace(trace)

    """
    fig.update_layout(updatemenus=[go.layout.Updatemenu(
                        active=0,
                        buttons=buttonlist)],
                      scene=dict(
                        xaxis=dict(
                            title_text="PCA 1"),
                        yaxis=dict(
                            title_text="PCA 2"),
                        zaxis=dict(
                            title_text="PCA 3"),
                        )
                     )
    """
    fig.update_layout(updatemenus=[go.layout.Updatemenu(
        active=0,
        buttons=buttonlist)],
        scene=dict(
            xaxis=dict(
                title="",
                showticklabels=False),
            yaxis=dict(
                title="",
                showticklabels=False),
            zaxis=dict(
                title="",
                showticklabels=False)
        )
    )
    return fig


def pipeline_for_multiple_traces(words, n, model):
    tracelist = []
    for word in words:
        nn_dict = nearest_neighbours(word, n, model)

        df = pd.DataFrame.from_dict(nn_dict, orient="index")
        vocab = [x for x in df.index]

        coordinates = sk_PCA(df)

        new_dict = {}
        for w, v in zip(vocab, coordinates):
            new_dict[w] = np.real(v)

        df = pd.DataFrame.from_dict(new_dict, orient="index")
        tracelist.append(return_trace(df, word))

    fig = figure_with_multiple_traces(tracelist)
    print("Displaying visualisation in browser...")
    fig.show()

    # fig.write_html(f"{'_'.join([x for x in words])}.html", include_plotlyjs=False)


def we_pca_vis(we_model, dimensions):
    global dims
    dims = dimensions
    print("Loading model ...")
    model = KeyedVectors.load_word2vec_format(we_model, binary=False)
    print("Model loaded, computing visualisation ...")
    # Hier kannst du die Wörter der Word-Embeddings eingeben, die dich interessieren und die Anzahl der
    # Nearest Neighbors, die gezeigt werden sollen
    pipeline_for_multiple_traces(["Mann", "Frau", "Königin"], 25, model)
    print("Done")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Visualisiert Word Embeddings")
    parser.add_argument("we_model", type=str, help="Input-Datei: Word Embeddings Modell im word_vectors Format")
    parser.add_argument("dimensions", type=str, help="Dimensionen: '2d' oder '3d', je nachdem wie viele Dimensionen die Darstellung haben soll")
    args = parser.parse_args()
    we_pca_vis(args.we_model, args.dimensions)